	;;
	;; Accel.asm
	;;
	;; Handles all aspects of taking an acceleration sensor reading and
	;; conversion to engineering units. This includes a recursive filter
	;; used during pre-launch to determine the 1G acceleration offset.
	;;
	;; On exit, acceleration will be stored in T0-T3
	;;
	include p16f628.inc
	include data.inc
	include eedata.inc
	include adc.inc

	global	get_acceleration
	extern	eeread, FXM3216S



	code
	
	;; a_filter:	
	;;
	;; Input: Latest reading in T0-T1 (unchanged by code)
	;; 
	;; Perform recursive filter on acceleration data.
	;; Data is stored as 16.16 fixed point value. Filter constant
	;; is 256. Complicated by fact filtered values are in data page 1 so
	;; use AARG and BARG as temporary storage.
	;;
	;; I recently located a reference that had a simple equation for
	;; computing the 3dB corner frequency for this type of filter.
	;; The result for this filter at 64 SPS is 1/25Hz. This is certainly
	;; too agressive and possibly should be changed. But it would
	;; significantly complicate the implementation.
	;; Because the Kalman filter algorithm is very tolerant of offset
	;; errors, it will stay as it is for now.
	;;
	;; Start by subtracting (old value >> 8) from new (value << 8)
a_filter:
	movlw	AVE_ACCEL0B0	; setup FSR pointer for copy
	movwf	FSR
	bcf	STATUS,IRP
	
	movf	INDF,W		; get copy of average shifted right 8 bits
	movwf	AARGB1
	incf	FSR,F
	movf	INDF,W
	movwf	AARGB2
	incf	FSR,F
	movf	INDF,W
	movwf	AARGB3
	incf	FSR,F
	clrf	AARGB0

	movf	T0,W		; make copy of new value and make it 32 bits
	movwf	BARGB1		; shifted left 8 bits
	movf	T1,W
	movwf	BARGB2
	clrf	BARGB0
	clrf	BARGB3

	movf	AARGB3,W	; subtract last average value from current
	subwf	BARGB3,F	; reading

	movf	AARGB2,W
	btfss	STATUS,C
	incfsz	AARGB2,W
	subwf	BARGB2,F
	
	movf	AARGB1,W
	btfss	STATUS,C
	incfsz	AARGB1,W
	subwf	BARGB1,F

	movf	AARGB0,W
	btfss	STATUS,C
	incfsz	AARGB0,W
	subwf	BARGB0,F

	;; Now add the result back into average value

	movf	BARGB3,W
	addwf	INDF,F
	decf	FSR,F

	movf	BARGB2,W
	btfsc	STATUS,C
	incfsz	BARGB2,W
	addwf	INDF,F
	decf	FSR,F
	
	movf	BARGB1,W
	btfsc	STATUS,C
	incfsz	BARGB1,W
	addwf	INDF,F
	decf	FSR,F
	
	movf	BARGB0,W
	btfsc	STATUS,C
	incfsz	BARGB0,W
	addwf	INDF,F

	return
	
get_acceleration:	
	;;
	;; First things first. Read a value from the ADC
	;;
	movlw	SECH0		; channel 0 has the accelerometer
	movwf	T0
	call	readADC		; result is in T0,T1
	movf	T0,W
	movwf	RAWACCELB0
	movf	T1,W
	movwf	RAWACCELB1

	call	a_filter	; update recursive filtered value

	;;
	;; Now remove the 1G offset by subtracting from long term average on pad
	;; which is in AVE_ACCEL2B0-3. First copy AVE_ACCEL2 from data page 2
	;;
	movlw	AVE_ACCEL2B0	; setup FSR pointer for copy
	movwf	FSR
	bcf	STATUS,IRP
	
	movf	INDF,W		; copy data
	movwf	AARGB0
	incf	FSR,F
	movf	INDF,W
	movwf	AARGB1
	incf	FSR,F
	movf	INDF,W
	movwf	AARGB2
	incf	FSR,F
	movf	INDF,W
	movwf	AARGB3

	

	;; Do the subtraction now

	movf	T1,W
	subwf	AARGB1,F

	movf	T0,W
	btfss	STATUS,C
	incfsz	T0,W
	subwf	AARGB0,F


	;;
	;; Now get the sensitivity (meters/sec/sec per count) for the multiply
	;; The sensitivity is a 0.16 fixed point fraction. It MUST have a
	;; magnitude less than 1/2.
	;;

	movlw	ACCEL_SENS_HI
	call	eeread
	movwf	BARGB0
	movlw	ACCEL_SENS_LO
	call	eeread
	movwf	BARGB1
	

	call	FXM3216S	; 32X16 signed multiply

	;; Copy useful bits to T0-T3

	movf	AARGB0,W
	movwf	T0
	movf	AARGB1,W
	movwf	T1
	movf	AARGB2,W
	movwf	T2
	movf	AARGB3,W
	movwf	T3

	return
	end
