	;;
	;; Altitude.asm
	;;
	;; Handles all aspects of taking a pressure sensor reading and
	;; conversion to altitude.
	;;
	;; On exit, altitude will be stored in T0-T3
	;;
	include p16f628.inc
	include data.inc
	include eedata.inc
	include adc.inc

	global	get_altitude
	extern	eeread, FXM3216S

		;; The minimum usable output voltage according to the MPXA4115A
	;; data sheet is 0.2 volts. This translates to 163 ADC counts.
	;; Actual data shows that the sensor output clips at 68 mV but it
	;; has higher non-linearity errors in this region. Therefore 200mV
	;; is used as the cutoff.
	
MINPRESSURE	equ	.163	; Minimum usable pressure sensor reading


	code
	
get_altitude:
	;;
	;; First things first. Read a value from the ADC
	;;
	movlw	SECH1		; channel 1 has the pressure sensor
	movwf	T0
	call	readADC		; result is in T0,T1
	movf	T0,W		; copy
	movwf	PRESSUREB0
	movf	T1,W
	movwf	PRESSUREB1
	;; Read 8.8 fixed point slope from eeprom and multiply by
	;; least significant 8 bits of pressure reading
	;; Since ADC result is 12 bits, T0 holds the table index and
	;; T1 the multiplicand
	movf	T0,W		; get index
	addlw	PSLOPE_HI
	call	eeread		; read from eeprom
	movwf	AARGB1		; set up for multiply
	movf	T0,W		; get index
	addlw	PSLOPE_LO
	call	eeread		; read from eeprom
	movwf	AARGB2		; set up for multiply
	clrf	AARGB3		; clear LSB
	movlw	0
	btfsc	AARGB1,7	; see if negative
	movlw	0xff		; make all ones for..
	movwf	AARGB0		; sign extend
	clrf	BARGB0		; fix up second argument
	movf	T1,W
	movwf	BARGB1
	call	FXM3216S	; do multiply
	;;
	;; Now we add in the base address
	;;
	movf	T0,W		; index into data
	addlw	PBASE_LO
	call	eeread
	addwf	AARGB3,F
	btfsc	STATUS,C	; handle carry
	incf	AARGB2,F
	movf	T0,W		; index into data for upper byte
	addlw	PBASE_HI
	call	eeread
	addwf	AARGB2,F		; and add it in
	;;
	;; Finally, copy result to T0-T3
	movf	AARGB2,W
	movwf	T0
	movf	AARGB3,W
	movwf	T1
	movf	AARGB4,W
	movwf	T2
	movf	AARGB5,W
	movwf	T3


	;;
	;; Check for out of range sensor (altitude too high)
	;;
	bcf	STATUS,C
	movf	PRESSUREB0,W	; first check upper byte to see if zero
	btfss	STATUS,Z
	goto	p_ok
	movlw	MINPRESSURE	; upper byte was zero so check lower
	subwf	PRESSUREB1,W
	btfsc	STATUS,C
	goto	p_ok
	;;
	;; Pressure reading is too low. Set carry flag and return
	bsf	STATUS,C
	return
	;;
	;; Pressure is OK
p_ok:	bcf	STATUS,C	; indicate valid altitude
	return
	end
