/*

 
 NAME
        kgain.c -  Compute Kalman gains for single
	measurement constant acceleration model.
                                                    
 SYNOPSIS                                           
        kgain <sample rate> <measurement noise> <measurement/model noise ratio>
                                                    
 DESCRIPTION:                                       
        Computes Kalman gains.

        Input is:
	sample rate in samples per second (integer)
	measurement noise (standard deviation)
	ratio of measurement noise to model noise

	Output is:
	K1 = <gain>  (pressure)
	K2 = <gain>  (velocity)
	K3 = <gain>  (acceleration)

                                                        
 AUTHOR                                                 
	  Hacked by David Schultz

                                                        
 REPORTING BUGS AND REVISIONS                           
        Report bugs and revisions to                    
	david.schultz@earthlink.net
                                                        
                                                        
 COMPILING                                              
   cc -o kgain.exe kgain.c
                                                        
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  MEASUREMENTSIGMA	15.0
#define  MODELSIGMA		0.05

double measurement_variance = MEASUREMENTSIGMA*MEASUREMENTSIGMA;
double model_variance  = MODELSIGMA*MODELSIGMA;

main(int argc, char ** argv)
{
	char	buf[512];

	int     i, k, notdone;

	double	pest[3][3]    = { 2, 0, 0,
				  0, 9, 0,
				  0, 0, 9 };
	double  pestp[3][3]   = { 0, 0, 0,
				   0, 0, 0,
				   0, 0, 0 };
	double	phi[3][3]     = { 1, 0, 0,
				 0, 1, 0,
				 0, 0, 1.0 };
	double  phit[3][3]    = { 1, 0, 0,
				  0, 1, 0,
				  0, 0, 1.0 };
	double  gain[3] = { 0.01, 0.01, 0.01 };
	double  lastgain[3] = {0.0, 0.0, 0.0 };
	double  dt;
	double	term[3][3];

	k = 0;

	/* read input parameters */
	if (argc != 4)
	{
	  printf("Usage: kgain <sample rate> <measurement noise> <measurement/model noise ratio>\n");
	  exit(1);
	}
	else
	  {
		double temp;
		dt = 1/atof(argv[1]);
		temp = atof(argv[2]);
		measurement_variance = temp * temp;
		temp = atof(argv[3]);
		model_variance = measurement_variance/temp;
	}
	
/* Initialize */


/*
 Fill in state transition matrix and its transpose
*/
      phi[0][1]    = dt;
      phi[1][2]    = dt;
      phi[0][2]    = dt*dt/2.0;
      phit[1][0]    = dt;
      phit[2][1]    = dt;
      phit[2][0]    = dt*dt/2.0;


	while( 1 )
	{


		/* Propagate state covariance */

         term[0][0] = phi[0][0] * pest[0][0] + phi[0][1] * pest[1][0] + phi[0][2] * pest[2][0];
         term[0][1] = phi[0][0] * pest[0][1] + phi[0][1] * pest[1][1] + phi[0][2] * pest[2][1];
         term[0][2] = phi[0][0] * pest[0][2] + phi[0][1] * pest[1][2] + phi[0][2] * pest[2][2];
         term[1][0] = phi[1][0] * pest[0][0] + phi[1][1] * pest[1][0] + phi[1][2] * pest[2][0];
         term[1][1] = phi[1][0] * pest[0][1] + phi[1][1] * pest[1][1] + phi[1][2] * pest[2][1];
         term[1][2] = phi[1][0] * pest[0][2] + phi[1][1] * pest[1][2] + phi[1][2] * pest[2][2];
         term[2][0] = phi[2][0] * pest[0][0] + phi[2][1] * pest[1][0] + phi[2][2] * pest[2][0];
         term[2][1] = phi[2][0] * pest[0][1] + phi[2][1] * pest[1][1] + phi[2][2] * pest[2][1];
         term[2][2] = phi[2][0] * pest[0][2] + phi[2][1] * pest[1][2] + phi[2][2] * pest[2][2];

         pestp[0][0] = term[0][0] * phit[0][0] + term[0][1] * phit[1][0] + term[0][2] * phit[2][0];
         pestp[0][1] = term[0][0] * phit[0][1] + term[0][1] * phit[1][1] + term[0][2] * phit[2][1];
         pestp[0][2] = term[0][0] * phit[0][2] + term[0][1] * phit[1][2] + term[0][2] * phit[2][2];
         pestp[1][0] = term[1][0] * phit[0][0] + term[1][1] * phit[1][0] + term[1][2] * phit[2][0];
         pestp[1][1] = term[1][0] * phit[0][1] + term[1][1] * phit[1][1] + term[1][2] * phit[2][1];
         pestp[1][2] = term[1][0] * phit[0][2] + term[1][1] * phit[1][2] + term[1][2] * phit[2][2];
         pestp[2][0] = term[2][0] * phit[0][0] + term[2][1] * phit[1][0] + term[2][2] * phit[2][0];
         pestp[2][1] = term[2][0] * phit[0][1] + term[2][1] * phit[1][1] + term[2][2] * phit[2][1];
         pestp[2][2] = term[2][0] * phit[0][2] + term[2][1] * phit[1][2] + term[2][2] * phit[2][2];

         pestp[2][2] = pestp[2][2] + model_variance;
/*
 Calculate Kalman Gain
*/
         gain[0] = pestp[0][0] / (pestp[0][0]  + measurement_variance);
		   
         gain[1] =  pestp[1][0]  / (pestp[0][0] + measurement_variance);

         gain[2] =  pestp[2][0]  / (pestp[0][0] + measurement_variance);

	 /*
	   Propagate state covariance
	 */

         pest[0][0] = pestp[0][0] * (1.0 - gain[0]);
         pest[0][1] = pestp[1][0] * (1.0 - gain[0]);
	 pest[0][2] = pestp[2][0] * (1.0 - gain[0]);
         pest[1][0] = pestp[0][1] - gain[1] * pestp[0][0];
         pest[1][1] = pestp[1][1] - gain[1] * pestp[1][0];
	 pest[1][2] = pestp[2][1] - gain[1] * pestp[2][0];
         pest[2][0] = pestp[0][2] - gain[2] * pestp[0][0];
         pest[2][1] = pestp[1][2] - gain[2] * pestp[1][0];
	 pest[2][2] = pestp[2][2] - gain[2] * pestp[2][0];

	 /* Check for convergance. Criteria is less than .001% change from last
	  * time through the mill.
	  */
	 notdone = 0;
	 k++;
	 for( i = 0; i <= 2; i++)
	   {
	       if( (gain[i] - lastgain[i])/lastgain[i] > 0.00001)
		 notdone++;
	       lastgain[i] = gain[i];
	     }
	 if( notdone )
	   continue;
	 else
	   break;
	}
      printf("Input noise values used (standard deviation):\n");
      printf("#Altitude       - %15f \n", sqrt(measurement_variance));
      printf("#Model noise    - %15f \n#\n", sqrt(model_variance));
      printf("#Kalman gains converged after %d iterations.\n#", k);
      printf("\nK1 = %15.10lf\nK2 = %15.10lf\nK3 = %15.10lf\n", gain[0], gain[1], gain[2]);

      printf("16 bit fixed point gains (fraction only) in hex\n");
      printf("K1_HI = %x\n", (int)(gain[0] * 256));
      printf("K1_LO = %x\n", (int)(gain[0] * 256 * 256) & 255);
      printf("K2_HI = %x\n", (int)(gain[1] * 256));
      printf("K2_LO = %x\n", (int)(gain[1] * 256 * 256) & 255);
      printf("K3_HI = %x\n", (int)(gain[2] * 256));
      printf("K3_LO = %x\n", (int)(gain[2] * 256 * 256) & 255);

}
