/*

 
 NAME
        kgain32.c -  Compute Kalman gains for two 
	measurement constant acceleration model.
                                                    
 SYNOPSIS                                           
        kgain <sample rate> <accel noise> <ppressure noise> <accel measurement/model noise ratio>
                                                    
 DESCRIPTION:                                       
        Computes Kalman gains.

        Input is:
	sample rate in samples per second (integer)
	acceleration measurement noise (standard deviation)
	pressure measurement noise (standard deviation)
	ratio of measurement noise to model noise

	Output is:
	K1 = <gain>  (pressure)
	K2 = <gain>  (velocity)
	K3 = <gain>  (acceleration)
	K4
	K5
	K6
                                                        
 AUTHOR                                                 
	  Hacked by David Schultz

                                                        
 REPORTING BUGS AND REVISIONS                           
        Report bugs and revisions to                    
	david.schultz@earthlink.net
                                                        
                                                        
 COMPILING                                              
   cc -o kgain32.exe kgain32.c
                                                        
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* default values */

#define  ALTITUDESIGMA	       15.0
#define  ACCELERATIONSIGMA      6.0
#define  MODELSIGMA		0.6

double altitude_variance = ALTITUDESIGMA*ALTITUDESIGMA;
double acceleration_variance = ACCELERATIONSIGMA*ACCELERATIONSIGMA;
double model_variance  = MODELSIGMA*MODELSIGMA;

main(int argc, char ** argv)
{
	char	buf[512];

	int     i, j, k, notdone;

	double	pest[3][3]    = { 2, 0, 0,
				  0, 9, 0,
				  0, 0, 9 };
	double  pestp[3][3]   = { 0, 0, 0,
				   0, 0, 0,
				   0, 0, 0 };
	double	phi[3][3]     = { 1, 0, 0,
				 0, 1, 0,
				 0, 0, 1.0 };
	double  phit[3][3]    = { 1, 0, 0,
				  0, 1, 0,
				  0, 0, 1.0 };
	double  kgain[3][2], lastkgain[3][2];
	double  dt, det;
	double	term[3][3];

	k = 0;


	/* read input parameters */
	if (argc != 5)
	{
	  printf("Usage: kgain <sample rate> <accel noise> <pressure noise> <measurement/model noise ratio>\n");
	  exit(1);
	}
	else
	  {
		double temp;
		dt = 1/atof(argv[1]);
 		temp = atof(argv[2]);
		acceleration_variance = temp * temp;
 		temp = atof(argv[3]);
		altitude_variance = temp * temp;
		temp = atof(argv[4]);
		model_variance = acceleration_variance/temp;
	}
	
/* Initialize */

/*
 Fill in state transition matrix and its transpose
*/
      phi[0][1]    = dt;
      phi[1][2]    = dt;
      phi[0][2]    = dt*dt/2.0;
      phit[1][0]    = dt;
      phit[2][1]    = dt;
      phit[2][0]    = dt*dt/2.0;



      for( i = 0; i <= 2; i++)
	for( j = 0; j <=1; j++)
	  {
	    lastkgain[i][j] = .001;
	    kgain[i][j] = 1.0;
	  }
      k = 0;


      /* Compute the Kalman gain matrix. */
      while(1)
	{
	  /* Propagate state covariance */

	  term[0][0] = phi[0][0] * pest[0][0] + phi[0][1] * pest[1][0] + phi[0][2] * pest[2][0];
	  term[0][1] = phi[0][0] * pest[0][1] + phi[0][1] * pest[1][1] + phi[0][2] * pest[2][1];
	  term[0][2] = phi[0][0] * pest[0][2] + phi[0][1] * pest[1][2] + phi[0][2] * pest[2][2];
	  term[1][0] = phi[1][0] * pest[0][0] + phi[1][1] * pest[1][0] + phi[1][2] * pest[2][0];
	  term[1][1] = phi[1][0] * pest[0][1] + phi[1][1] * pest[1][1] + phi[1][2] * pest[2][1];
	  term[1][2] = phi[1][0] * pest[0][2] + phi[1][1] * pest[1][2] + phi[1][2] * pest[2][2];
	  term[2][0] = phi[2][0] * pest[0][0] + phi[2][1] * pest[1][0] + phi[2][2] * pest[2][0];
	  term[2][1] = phi[2][0] * pest[0][1] + phi[2][1] * pest[1][1] + phi[2][2] * pest[2][1];
	  term[2][2] = phi[2][0] * pest[0][2] + phi[2][1] * pest[1][2] + phi[2][2] * pest[2][2];

	  pestp[0][0] = term[0][0] * phit[0][0] + term[0][1] * phit[1][0] + term[0][2] * phit[2][0];
	  pestp[0][1] = term[0][0] * phit[0][1] + term[0][1] * phit[1][1] + term[0][2] * phit[2][1];
	  pestp[0][2] = term[0][0] * phit[0][2] + term[0][1] * phit[1][2] + term[0][2] * phit[2][2];
	  pestp[1][0] = term[1][0] * phit[0][0] + term[1][1] * phit[1][0] + term[1][2] * phit[2][0];
	  pestp[1][1] = term[1][0] * phit[0][1] + term[1][1] * phit[1][1] + term[1][2] * phit[2][1];
	  pestp[1][2] = term[1][0] * phit[0][2] + term[1][1] * phit[1][2] + term[1][2] * phit[2][2];
	  pestp[2][0] = term[2][0] * phit[0][0] + term[2][1] * phit[1][0] + term[2][2] * phit[2][0];
	  pestp[2][1] = term[2][0] * phit[0][1] + term[2][1] * phit[1][1] + term[2][2] * phit[2][1];
	  pestp[2][2] = term[2][0] * phit[0][2] + term[2][1] * phit[1][2] + term[2][2] * phit[2][2];

	  pestp[2][2] = pestp[2][2] + model_variance;
	 /*
	   Calculate Kalman Gain
	 */
	  det = (pestp[0][0]+altitude_variance)*(pestp[2][2] + acceleration_variance) - pestp[2][0] * pestp[0][2];


	 kgain[0][0] = (pestp[0][0] * (pestp[2][2] + acceleration_variance) - pestp[0][2] * pestp[2][0])/det;

	 kgain[0][1] = (pestp[0][0] * (-pestp[0][2]) + pestp[0][2] * (pestp[0][0] + altitude_variance))/det;

	 kgain[1][0] = (pestp[1][0] * (pestp[2][2] + acceleration_variance) - pestp[1][2] * pestp[2][0])/det;

	 kgain[1][1] = (pestp[1][0] * (-pestp[0][2]) + pestp[1][2] * (pestp[0][0] + altitude_variance))/det;

	 kgain[2][0] = (pestp[2][0] * (pestp[2][2] + acceleration_variance) - pestp[2][2] * pestp[2][0])/det;

	 kgain[2][1] = (pestp[2][0] * (-pestp[0][2]) + pestp[2][2] * (pestp[0][0] + altitude_variance))/det;

         pest[0][0] = pestp[0][0] * (1.0 - kgain[0][0]) - kgain[0][1]*pestp[2][0];
         pest[0][1] = pestp[0][1] * (1.0 - kgain[0][0]) - kgain[0][1]*pestp[2][1];
	 pest[0][2] = pestp[0][2] * (1.0 - kgain[0][0]) - kgain[0][1]*pestp[2][2];
         pest[1][0] = pestp[0][0] * (-kgain[1][0]) + pestp[1][0] - kgain[1][1]*pestp[2][0];
         pest[1][1] = pestp[0][1] * (-kgain[1][0]) + pestp[1][1] - kgain[1][1]*pestp[2][1];
	 pest[1][2] = pestp[0][2] * (-kgain[1][0]) + pestp[1][2] - kgain[1][1]*pestp[2][2];
         pest[2][0] = (1.0 - kgain[2][1]) * pestp[2][0] - kgain[2][0] * pestp[2][0];
         pest[2][1] = (1.0 - kgain[2][1]) * pestp[2][1] - kgain[2][0] * pestp[2][1];
	 pest[2][2] = (1.0 - kgain[2][1]) * pestp[2][2] - kgain[2][0] * pestp[2][2];
       
	 /* Check for convergance. Criteria is less than .001% change from last
	  * time through the mill. Because I have noticed very early convergance
	  * in a few cases (< 20 iterations) I am delaying this for at least 
	  * 100 iterations.
	  */
	 notdone = 0;
	 k++;
	 for( i = 0; i <= 2; i++)
	   for( j = 0; j <= 1; j++)
	     {
	       if( (kgain[i][j] - lastkgain[i][j])/lastkgain[i][j] > 0.00001)
		 notdone++;
	       lastkgain[i][j] = kgain[i][j];
	     }
	 if( notdone || k < 100)
	   continue;
	 else
	   break;
	}


      printf("Input noise values used (standard deviation):\n");
      printf("#Altitude       - %15f feet\n", sqrt(altitude_variance));
      printf("#Acceleration   - %15f feet/sec/sec\n", sqrt(acceleration_variance));
      printf("#Model noise    - %15f feet/sec/sec\n#\n", sqrt(model_variance));
      printf("Sample rate = %s\n\n", argv[1]);

      printf("#Estimated output first order statistics (standard deviation):\n");
      printf("#Altitude     - %15f feet\n", sqrt(pest[0][0]));
      printf("#Velocity     - %15f feet/sec\n", sqrt(pest[1][1]));
      printf("#Acceleration - %15f feet/sec/sec\n", sqrt(pest[2][2]));
      printf("#Kalman gains converged after %d iterations.\n#", k);
      for( i = 0; i <= 2; i++)
	for( j = 0; j <= 1; j++)
	  printf("%15f ", kgain[i][j]);
      printf("\n#\n");

      printf("0.16 bit fixed point gains (fraction only) in hex\n");
      printf("K1_HI = %x\n", (int)(kgain[0][0] * 256));
      printf("K1_LO = %x\n", (int)(kgain[0][0] * 256 * 256) & 255);
      printf("K2_HI = %x\n", (int)(kgain[1][0] * 256));
      printf("K2_LO = %x\n", (int)(kgain[1][0] * 256 * 256) & 255);
      printf("K3_HI = %x\n", (int)(kgain[2][0] * 256));
      printf("K3_LO = %x\n", (int)(kgain[2][0] * 256 * 256) & 255);
      printf("K4_HI = %x\n", (int)(kgain[0][1] * 256));
      printf("K4_LO = %x\n", (int)(kgain[0][1] * 256 * 256) & 255);
      printf("K5_HI = %x\n", (int)(kgain[1][1] * 256));
      printf("K5_LO = %x\n", (int)(kgain[1][1] * 256 * 256) & 255);
      printf("K6_HI = %x\n", (int)(kgain[2][1] * 256));
      printf("K6_LO = %x\n", (int)(kgain[2][1] * 256 * 256) & 255);

}
