	include p16f628.inc
	include data.inc
	include adc.inc
	include eedata.inc

	global	KALMAN_STEP,kalman_init,set_alt_gains
	extern	FXM3216S, get_altitude, get_acceleration, eeread
	extern	beep_alt

	;; Copyright 2002 David W. Schultz
	;; 
	;; Quick and dirty version of Kalman filter for PIC16 processors.
	;; Uses 32X16 multiply routine extracted from Microchip application
	;; note.
	;;
	;; This version incorporates both pressure and acceleration
	;; measurements. It uses ~330 words of code space. Execution time
	;; is dominated by the 32x16 bit multiplies which require ~400 machine
	;; cycles each. With 6 multiplies that means 2400 machine cycles.
	;; Double that to allow for everything else that goes on and you have
	;; a requirement for ~5000 machine cycles per sample. Since the PIC
	;; has four clock cycles per machine cycle that it 20,000 clocks. A
	;; 4MHz clock would allow a maximum of 200 samples/second. So there is
	;; plenty of margin when running at 128 samples/second. 
	;;
	;; The primary limitation is code and data space availability. The
	;; 16F84 has 1024 words code space and 68 bytes data space. This code
	;; takes 309 words of code space. Could be pretty
	;; tight. The 16F628 which has 2K of code space would be a better
	;; choice. Especially since it is pin compatible with the 16F84, has
	;; more features, more code space, more data space, and is cheaper.
	;;
	;; Changing the sample rate requires changing the number of bits
	;; to shift in the state update code. It is currently set for
	;; 128 SPS.
	;;
	;; Kalman gains are computed by a "C" application and are determined
	;; by sample rate, measurement noise, and model noise. To compute
	;; gains, select sample rate, measure sensor noise
	;; (standard deviation), and select measurement/model noise ratio. I
	;; recomend a ratio of 10 to 100. Input parameters into kgain32.exe and
	;; note results. Gains shown here are based on 64 SPS, noise of 15'
	;; for pressure altitude, 6ft/sec/sec for acceleration, and  a noise
	;; ratio of 100		
	;;
	;; Initialization:
	;;
	;; Clear VELB0-VELB3 and ACCELB0-ACCELB3. Read altitude value and load
	;; into ALTB0-3. Then you can start running the filter.
	;;
	;; If you start the filter with random garbage for the intial values
	;; you will get random garbage out for quite a while.
	;;
	;; A note on number representation.
	;;
	;; Altitude, velocity, and acceleration are stored in 16.16 fixed
	;; point format. The two most significant bytes represent the integer
	;; portion of the number and the two LSB's the fractional portion. It's
	;; just like base ten if you are missing 8 fingers. :-)
	;;
	;; The Kalman gains are only 16 bit values because the integer part is
	;; zero. This allows use of a 32X16 multiply with 48 bit result instead
	;; of a 32x32 bit multiply with 64 bit result. It shouldn't be a
	;; problem but there is a chance that at very low sample rate, one
	;; of the gains could be larger than one. You have two choices:	1)
	;; increase sample rate. 2) Modify number representation and code.
	;; 
	;; 



	code

	;; Initialization of the Kalman filter is pretty simple. Just set
	;; the velocity and acceleration to zero. Then get the current altitude
	;; from the pressure sensor to set altitude.
	;; 
kalman_init:
	clrf	VELB0
	clrf	VELB1
	clrf	VELB2
	clrf	VELB3
	clrf	ACCELB0
	clrf	ACCELB1
	clrf	ACCELB2
	clrf	ACCELB3


	;; Read current altitude

	call	get_altitude	; result is in T0-3

	movf	T0,W		; copy to altitude state variable
	movwf	ALTB0
	movf	T1,W
	movwf	ALTB1
	movf	T2,W
	movwf	ALTB2
	movf	T3,W
	movwf	ALTB3
	;;
	;; Initialize accelerometer averages
	;;
	movlw	SECH0		; read value from ADC
	movwf	T0
	call	readADC


	movf	T0,W
	bsf	STATUS,RP0	; date page 1
	movwf	AVE_ACCEL0B0
	movwf	AVE_ACCEL1B0
	movwf	AVE_ACCEL2B0
	bcf	STATUS,RP0	; data page 0
	movf	T1,W
	bsf	STATUS,RP0	; date page 1
	movwf	AVE_ACCEL0B1
	movwf	AVE_ACCEL1B1
	movwf	AVE_ACCEL2B1
	clrf	AVE_ACCEL0B2
	clrf	AVE_ACCEL1B2
	clrf	AVE_ACCEL2B2
	clrf	AVE_ACCEL0B3
	clrf	AVE_ACCEL1B3
	clrf	AVE_ACCEL2B3
	bcf	STATUS,RP0	; data page 0

	;; Set Kalman gains

	movlw	K1_HI
	call	eeread
	movwf	K1B0
	movlw	K1_LO
	call	eeread
	movwf	K1B1
	
	movlw	K2_HI
	call	eeread
	movwf	K2B0
	movlw	K2_LO
	call	eeread
	movwf	K2B1
	
	movlw	K3_HI
	call	eeread
	movwf	K3B0
	movlw	K3_LO
	call	eeread
	movwf	K3B1

	movlw	K4_HI
	call	eeread
	movwf	K4B0
	movlw	K4_LO
	call	eeread
	movwf	K4B1

	movlw	K5_HI
	call	eeread
	movwf	K5B0
	movlw	K5_LO
	call	eeread
	movwf	K5B1

	movlw	K6_HI
	call	eeread
	movwf	K6B0
	movlw	K6_LO
	call	eeread
	movwf	K6B1

	;; set timer tick to zero
	clrf	TICKB0
	clrf	TICKB1
	clrf	TICKB2

	return

;;
;;	Set Kalman gains to the alternate, pressure sensor only, gain set.
;;	This is called at apogee since the orientation of the acclerometer 
;;	becomes pretty random at that point.
;;
	
set_alt_gains:

	movlw	ALTK1_HI
	call	eeread
	movwf	K1B0
	movlw	ALTK1_LO
	call	eeread
	movwf	K1B1
	
	movlw	ALTK2_HI
	call	eeread
	movwf	K2B0
	movlw	ALTK2_LO
	call	eeread
	movwf	K2B1
	
	movlw	ALTK3_HI
	call	eeread
	movwf	K3B0
	movlw	ALTK3_LO
	call	eeread
	movwf	K3B1

	clrf	K4B0
	clrf	K4B1
	clrf	K5B0
	clrf	K5B1
	clrf	K6B0
	clrf	K6B1

	return


	;;  Multiply/accumulate code. This routine is used by the state
	;;  correction code.
	;;
	;;  Before calling, T0-T3 should contain the inovation. BARG should
	;;  be loaded with the Kalman gain, and FSR should point to the
	;;  location to add the result to.
	;;
	;; NOTE: MSB of multi-precision numbers is in xxxB0 which has the
	;; lowest address. FSR should point to the LSB (xxxB3) of the target.
	;;
	;; 
KALMAC:	movf	T0,W		; Copy T0-T3 to AARG in preperation for multiply
	movwf	AARGB0
	movf	T1,W
	movwf	AARGB1
	movf	T2,W
	movwf	AARGB2
	movf	T3,W
	movwf	AARGB3

	call	FXM3216S	; Perform the signed multiply

	;; crap! I hate the fact that this stupid chip doesn't have an add
	;; with carry
	
	movf	AARGB3,W	; add upper 32 bits of result to altitude
	addwf	INDF,F
	
	decf	FSR,F		; point at next byte
	movf	AARGB2,W
	btfsc	STATUS,C	; handle the carry
	incfsz	AARGB2,W
	addwf	INDF,F
	
	decf	FSR,F
	movf	AARGB1,W
	btfsc	STATUS,C
	incfsz	AARGB1,W
	addwf	INDF,F
	
	decf	FSR,F
	movf	AARGB0,W
	btfsc	STATUS,C
	incfsz	AARGB0,W
	addwf	INDF,F
	retlw	0


	

	;;
	;; 
	;; Arithmetic right shift 32 bit T0-T3, LOOPCOUNT places
	;; 
RSTEMP:	movf	T0,W		; Copy MSB to use for sign extension
	movwf	SIGN
RSLOOP:	rlf	SIGN,W		; Pick up the sign bit in C
	rrf	T0,F
	rrf	T1,F
	rrf	T2,F
	rrf	T3,F
	decfsz	LOOPCOUNT,F
	goto	RSLOOP
	retlw	0
	
	;;
	;; Begin state update
	;;
	;; Start by computing Xt+1 = X + V*T + A*T*T/2
	;;
	;; Since T is a fractional power of two, we do this by shifting bits.
	;; T is 1/128
	;;
	;; First make a temporary copy of the last velocity estimate and
	;; shift it right by 8 bits
	;; 
KALMAN_STEP:
	movf	VELB2,W
	movwf	T3
	movf	VELB1,W
	movwf	T2
	movf	VELB0,W
	movwf	T1

	movlw	0				; sign extension
	btfsc	T1,7
	movlw	0xff
	movwf	T0

	rlf		VELB3,W		; pick up the msb of VELB3 for the left shift
	rlf		T3,F
	rlf		T2,F
	rlf		T1,F
;	rlf		T0,F		; redundant because of the sign extension


	;; Now add it to the last altitude estimate (32 bit add)
	;; 
	movf	T3,W		; add temp to altitude
	addwf	ALTB3,F

	movf	T2,W
	btfsc	STATUS,C	; handle the carry
	incfsz	T2,W
	addwf	ALTB2,F

	movf	T1,W
	btfsc	STATUS,C
	incfsz	T1,W
	addwf	ALTB1,F

	movf	T0,W
	btfsc	STATUS,C
	incfsz	T0,W
	addwf	ALTB0,F

	;; Now make a temporary copy of the last acceleration estimate
	;; and divide by 128*128*2 or shift right 15 bits. This is done
	;; in two stages. First as we are copying to a temp location, do a byte
	;; shift. Then since this shifts by 16 bits, shift left by one bit.
	;; This optimizes the shift pretty well but is dificult to change for 
	;; different sample rates
	;; 
	movf	ACCELB1,W
	movwf	T3
	movf	ACCELB0,W
	movwf	T2

	movlw	0				; sign extension
	btfsc	T2,7
	movlw	0xff
	movwf	T1
	movwf	T0

	rlf		ACCELB2,W		; pick up the msb of ACCELB2 for the left shift
	rlf		T3,F
	rlf		T2,F
;	rlf		T1,F			; redundant because of the sign extension


	;; Now add it into the last altitude estimate (32 bit add)
	;; 
	movf	T3,W		; add temp to altitude
	addwf	ALTB3,F

	movf	T2,W
	btfsc	STATUS,C	; handle the carry
	incfsz	T2,W
	addwf	ALTB2,F

	movf	T1,W
	btfsc	STATUS,C
	incfsz	T1,W
	addwf	ALTB1,F

	movf	T0,W
	btfsc	STATUS,C
	incfsz	T0,W
	addwf	ALTB0,F

	;; State update of altitude is now complete
	;;
	;; Now we update the velocity estimate
	;;
	;; Make a temporary copy of the last acceleration estimate and
	;; shift right 7 bits. Optimized.
	;; 
	movf	ACCELB2,W
	movwf	T3
	movf	ACCELB1,W
	movwf	T2
	movf	ACCELB0,W
	movwf	T1

	movlw	0				; sign extension
	btfsc	T1,7
	movlw	0xff
	movwf	T0

	rlf		ACCELB3,W		; pick up the msb of VELB3 for the left shift
	rlf		T3,F
	rlf		T2,F
	rlf		T1,F
;	rlf		T0,F		; redundant because of the sign extension


	;; Now add it to the velocity estimate
	;; 
	movf	T3,W		; add temp to altitude
	addwf	VELB3,F

	movf	T2,W
	btfsc	STATUS,C	; handle the carry
	incfsz	T2,W
	addwf	VELB2,F

	movf	T1,W
	btfsc	STATUS,C
	incfsz	T1,W
	addwf	VELB1,F

	movf	T0,W
	btfsc	STATUS,C
	incfsz	T0,W
	addwf	VELB0,F


	;; This completes the state update for velocity
	;;
	;; Acceleration is assumed to be constant so no update of it
	;;  is required. Move on to the state correction.
	;; 
	;; Begin state correction
	;; A miracle occurs and the latest measured altitude (in meters)
	;; appears in T0-T3 as a 16.16 fixed point number.

	call	get_altitude	; This takes care of the miracle
	
	;; 
	;; First up is a sanity check. The code that reads the pressure
	;; should set the carry flag if the range of the sensor is exceeded.
	;; This is best checked by comparing the ADC value so we don't do it
	;; here.
	;; Therefore, we check the carry flag and if it is set, we skip all of
	;; the pressure altitude inovation code.

	btfsc	STATUS,C
	goto	A_INOVAT	;  carry set, move on to altitude inovation
	

	;; Now we check to see if we are in the Mach transition region.
	;; Mach 1 = ~331m/s

	movlw	#1		; MSB should be = 1
	xorwf	VELB0,W
	btfss	STATUS,Z
	goto	P_INOVAT	; less than 256 or greater than 512 m/s, so go on
	;; Now we know that our velocity is in the range of 256 to 512 m/s
	;; An arbitrary range of Mach 1 +/- 75m/s is selected as the criteria.
	;;  This allows us to complete the test with only one more comparison
	;; since 331 m/s - 75 m/s = 256 m/s. So we just check VELB1 to see if
	;; it is greater than 150.

	movf	VELB1,W
	addlw	(0x100 - 0x96)
	btfsc	STATUS,C	;
	goto	P_INOVAT	; result was positive so carry on
	
	;; Copy measured pressure to estimate and then skip to acceleration
	;; inovation part of code

	movf	T0,W
	movwf	ALTB0
	movf	T1,W
	movwf	ALTB1
	movf	T2,W
	movwf	ALTB2
	movf	T3,W
	movwf	ALTB3
	goto	A_INOVAT

	;; Pressure inovation correction
	;; 
	;; First step is to subtract altitude from pressure reading to
	;; get the inovation. Since we need this value for three computations,
	;; it is left in T0-T3 for later use. Do not mess with T0-T3!

P_INOVAT:	
	movf	ALTB3,W
	subwf	T3,F

	movf	ALTB2,W
	btfss	STATUS,C
	incfsz	ALTB2,W
	subwf	T2,F

	movf	ALTB1,W
	btfss	STATUS,C
	incfsz	ALTB1,W
	subwf	T1,F

	movf	ALTB0,W
	btfss	STATUS,C
	incfsz	ALTB0,W
	subwf	T0,F

	;; Altitude correction
	;; Multiply the inovation (currently in T0-T3) by the appropriate
	;; Kalman gain and add the altitude estimate
	;; This is done by the KALMAC subroutine which copies to invation
	;; from T0-T3 to AARG and then calls the 32X16 multiply routine. It
	;; then adds the upeer 32 bits of the result to the altitude estimate.
	;; We use the upper 32 bits because we have multiplied two number with
	;; 16 bits each to the right of the binary point. Therefore the result
	;; has 32 bits to the right of the binary point and we are dropping
	;; the least significant 16 bits of the fractional part.
	;; 
			
	movf	K1B0,W		;  Load Kalman gain
	movwf	BARGB0
	movf	K1B1,W
	movwf	BARGB1

	movlw	ALTB3
	movwf	FSR
	
	call	KALMAC		; Perform the multiply/accumulate

	;; Velocity correction

	movf	K2B0,W		;  Load Kalman gain
	movwf	BARGB0
	movf	K2B1,W
	movwf	BARGB1

	movlw	VELB3
	movwf	FSR
	
	call	KALMAC		; Perform multiply/accumulate

	;;  Acceleration correction

	movf	K3B0,W		; Load Kalman gain
	movwf	BARGB0
	movf	K3B1,W
	movwf	BARGB1

	movlw	ACCELB3
	movwf	FSR

	call	KALMAC

	;; Once again a miracle occurs. This time the current acceleration
	;; measurement (In m/s/s) appears in T0-T3. This value must be offset
	;; so that while at rest on the pad, the acceleration is ZERO. It
	;; is very important that this value be as close to zero as possible.
	;; Fractional bits are really good so don't throw them away.
	;;
	;; This is just like the pressure inovation code except that we use
	;; different gains.
	
	;; First step is to subtract acceleration from measurement to
	;; get the inovation. This is left in T0-T3 for later use.

A_INOVAT:	
	call	get_acceleration ; handle the miracle bit

	movf	ACCELB3,W
	subwf	T3,F

	movf	ACCELB2,W
	btfss	STATUS,C
	incfsz	ACCELB2,W
	subwf	T2,F

	movf	ACCELB1,W
	btfss	STATUS,C
	incfsz	ACCELB1,W
	subwf	T1,F

	movf	ACCELB0,W
	btfss	STATUS,C
	incfsz	ACCELB0,W
	subwf	T0,F

	;; Altitude correction
			
	movf	K4B0,W		;  Load Kalman gain
	movwf	BARGB0
	movf	K4B1,W
	movwf	BARGB1

	movlw	ALTB3
	movwf	FSR
	
	call	KALMAC		; Perform the multiply/accumulate

	;; Velocity correction

	movf	K5B0,W		;  Load Kalman gain
	movwf	BARGB0
	movf	K5B1,W
	movwf	BARGB1

	movlw	VELB3
	movwf	FSR
	
	call	KALMAC		; Perform multiply/accumulate

	;;  Acceleration correction

	movf	K6B0,W		; Load Kalman gain
	movwf	BARGB0
	movf	K6B1,W
	movwf	BARGB1

	movlw	ACCELB3
	movwf	FSR

	call	KALMAC
	
	;; Fine'
	;; Done
	;; Compleate
	return
	end
