#include <stdio.h>
#include <string.h>
#include <math.h>
/*
Read and process a data file from my Roctronics altimeter.

The file consists of lines with eight bytes of data per line.

read file from standard input and write to standard output

*/

#define SAMPLE_RATE 128.0


#define OFFSET    10.721         /* kPa    */
#define SENSITIVITY       45.0           /* mv/kPa */
#define A_SENSITIVITY     0.317769       /* m/s/s per count */
int process_sample(int i);

unsigned short int data[32768];

double P_alt,
  accel_1G = 0.0,
  m_accel,
  i_velocity = 0.0,
  i_altitude = 0.0;

int main(int argc, char **argv)
{
  int c, i, j, start;
  char cbuf[128];
  unsigned char temp[16];
 float time = 0.0;

 /* Skip the everything up to the first 'S' */

 while( (c = getc(stdin)) != 'S')
   ;
 ungetc(c, stdin);

 /* Read all of the data into memory. */

 i = 0;
 while( fgets(cbuf, 127, stdin) )
    {
      sscanf( cbuf, "S%*4x%2hx%2hx%2hx%2hx%2hx%2hx%2hx%2hx",
	      &data[i], &data[i+1], &data[i+2], &data[i+3],
	      &data[i+4], &data[i+5], &data[i+6], &data[i+7]);
      i += 8;
      if( i >= 32768 )
	break;
    }

 printf("#%d bytes read\n", i);

  /* Now look for the start of flight marker somewhere in the first
   2048 bytes of data. Start of flight can be seen because the flight state
   will change from 0 to 1.
  */

  i = 0;
  while(1)
    {

      if( (data[i+9] & 0xf0) == 0x10 )
	break;
      else
	{
	  i += 16;
	  if( i >= 2048 )
	    {
	      printf("Error processing file. No start of flight marker.\n");
	      i = -16;
	      break;
	    }
	}
    }
  /* OK, now i is set to just before where the data starts */

  printf("# %d\n", i);
  printf("Time    altitude    velocity    accel    pressure    rawaccel    state    voltage   P altitude  M accel    I velocity   I altitude\n");
  start = i;

  /* process data from start to end of first 2048 bytes */
  i += 16;
  while( i < 0x800)
    {
      process_sample(i);
      i += 16;
    }

  /* Now do the data from the beginning of the buffer. */
  i = 0;
  while( i <= start )
    {
      process_sample(i);
      i += 16;
    }


  /* now finish with the rest. Scanning for the end of data marker. */

  i = 0x800;
  while( i < 32768 )
    {
      if( process_sample(i) == 7 )
	printf("#\n# END OF FLIGHT #\n#\n");
      i += 16;
    }
      
}


float time = -1.0;

unsigned char last_tick;

union { 
  long int  ival;
  char      cval[4];
} uval;

int process_sample(int i)
{

  float altitude;
  float velocity;
  float accel;
  int   pressure;
  int   rawaccel;
  float   delta_t;

  unsigned short int voltage;
  unsigned char state, time_tick;
  

  /* assemble values from the data */

  uval.cval[3] = data[i++];
  uval.cval[2] = data[i++];
  uval.cval[1] = data[i++];
  uval.cval[0] = 0;

  altitude = ((float)uval.ival)/ 65536.0;

  uval.cval[3] = data[i++];
  uval.cval[2] = data[i++];
  uval.cval[1] = data[i++];
  uval.cval[0] = 0;

  velocity = ((float)uval.ival)/ 65536.0;

  uval.cval[3] = data[i++];
  uval.cval[2] = data[i++];
  uval.cval[1] = data[i++];
  uval.cval[0] = 0;

  accel = ((float)uval.ival)/ 65536.0;

  state = data[i] >> 4;

  pressure = ((data[i++] & 0xf) << 8);
  pressure |= (data[i++] & 0xff);

  rawaccel = ((data[i++] & 0x0f) << 8);
  rawaccel |= (data[i++] & 0xff);

  time_tick = data[i++];

  if ( time < 0.0 )
      time = 0.0;
  else
    {
      if( last_tick > time_tick )
	delta_t = ((float)(time_tick + (0x100 - last_tick)))/SAMPLE_RATE;
      else
	delta_t = ((float)(time_tick - last_tick))/ SAMPLE_RATE;
      time += delta_t;
    }


  last_tick = time_tick;

  /* Process the last two bytes of data as a 16 bit integer */

  voltage  = ((data[i++] & 0xff) << 8);
  voltage |= (data[i++] & 0xff);

  /* Do some additional computations on the raw sensor data. */

  
      P_alt = (288.14- 288.08 *
	       pow((((double)(pressure)/4096.0)*5000.0/SENSITIVITY + OFFSET)/101.29, 1/5.256))
	/0.00649;

      if( accel_1G == 0 )
	accel_1G = rawaccel;
      m_accel = (double)(accel_1G - rawaccel)*A_SENSITIVITY;
      i_velocity = i_velocity + m_accel * delta_t;
      i_altitude = i_altitude + i_velocity * delta_t;




  /* Now it is time to print out a line of data! */

  printf("%10f  %10f  %10f  %10f  %10d  %10d  %10d %10d %10f %10f %10f %10f\n",
	 time, altitude, velocity, accel, pressure, rawaccel, state, voltage,
	 P_alt, m_accel, i_velocity, i_altitude);

  return state;
}


