	include	p16f628.inc
	include data.inc
	
	extern	read_block, hex_to_ascii
	global	mother

	;; Serial port routines for 16F628
	;;
	;; This code is called on power up and it will look for
	;; someone on the other end of the serial port for a couple
	;; of seconds. If nothing responds, it is assumed that no one
	;; is there and control shifts to the regular altimter code.
	;;
	;; Therefore, once the altimeter is beeping out status, there is
	;; no way it will drop out of that mode.
	;;
	;; If someone is found on the other end of the serial port, we
	;; start a data dump.
	;; 
HOWDY	equ	'X'
HELLO	equ	'M'

	code
mother:	call	init		; set up serial port


	;;
	;; See if anyone is listening. Retry a maximum of 128 times
	;;
	movlw	0X80
	movwf	LOOPCOUNT
poll_loop:
	movf	RCREG,W		; make sure receive buffer is empty
	movf	RCREG,W
	movlw	HELLO		; try to get someone's attention
	call	txchar
	call	rx_char_timed	; see if there is a response
	xorlw	HOWDY		; make sure it is the correct response
	btfsc	STATUS,Z
	goto	got_something
	decfsz	LOOPCOUNT,F
	goto	poll_loop	; try again

mother_done:
	;;
	;; No body responded. Clean up after ourselves and return
	;;
	bcf	RCSTA,SPEN	; turn off receiver
	bsf	STATUS,RP0	; data page 1
	bcf	TXSTA,TXEN	; turn off transmitter
	bcf	STATUS,RP0	; page 0
	return			; failed, let the altimeter code run

	;; Somebody responded. Start the data dump.
	;;
	;; All data sent over the line is in the form of printable ASCII
	;; characters.
	;; Data is sent in 8 byte blocks. Each block begins with an ASCII 'S'
	;; a two byte block number and ends with a single byte checksum
	;; followed by a <CR><LF> pair.
	;; Then there is a wait for single acknowledge byte from the host.
	;; If the returned character is anything other than 'G' then the
	;; block is sent again. The block is also resent of there is no
	;; response from the host.
	;; 

got_something:
	
	clrf	EE_ADDRESSB0	; start at beginning
	clrf	EE_ADDRESSB1
dloop:	call	send_block	; send block of data
;	call	rx_char_timed	; get response
;	xorlw	'G'		; positive acknowledge?
;	btfss	STATUS,Z
;	goto	dloop
	movlw	8
	addwf	EE_ADDRESSB1,F	; increment address
	btfsc	STATUS,C
	incf	EE_ADDRESSB0,F
	btfss	EE_ADDRESSB0,7	; check if done
	goto	dloop
	
wloop:	goto	wloop		; infinite loop
	



	;;
	;; Send a block of data. Data is read from the EEPROM using the
	;; current address in EE_ADDRESS.
	;; 
send_block:	
	movlw	BUFP		; load address of buffer
	movwf	FSR
	bsf	STATUS,IRP
	clrf	TEMPB1		; used for checksum
	call	read_block	; read block of data from serial EEPROM
	movlw	'S'		; send start character
	call	txchar

	movf	EE_ADDRESSB0,W		; send block address
	call	send_byte
	movf	EE_ADDRESSB1,W
	call	send_byte

	movlw	8
	movwf	LOOPCOUNT
	movlw	BUFP		; start of buffer
	movwf	FSR
sloop:	movf	INDF,W		; get byte
	call	send_byte
	incf	FSR,F
	decfsz	LOOPCOUNT,F
	goto	sloop
	movf	TEMPB1,W	; send checksum
	call	send_byte
	movlw	0x0d
	call	txchar
	movlw	0x0a
	call	txchar
	return

	;; Convert the single byte in "W" to two ASCII characters and send.
send_byte:
	movwf	TEMPB0		; save it somewhere safe
	addwf	TEMPB1,F	; update checksum
	swapf	TEMPB0,W
	andlw	0x0f		; get upper nybble
	call	hex_to_ascii	; convert to ASCII
	call	txchar		; send it
	movf	TEMPB0,W
	andlw	0x0f		; get lower nybble
	call	hex_to_ascii
	call	txchar
	return

	;;
	;; We don't want to wait forever, so use the timer ticks to
	;; limit the wait. Wait a maximum of 4 ticks while polling to
	;; see if a character has arrived.
rx_char_timed:	
	movlw	4
	movwf	BEEP_TICK	; the beeper isn't using this now
rtloop:	btfsc	PIR1,RCIF	; check received character flag
	goto	r_arrived	; something is here
	btfss	TFLAG,0		; check for timer tick
	goto	rtloop		; repeat
	clrf	TFLAG
	decfsz	BEEP_TICK,F
	goto	rtloop
	;;
	;; SInce we are here, we have timed out waiting. Return a value
	;; of zero.
	;;
	retlw	0
r_arrived:
	movf	RCREG,W		; return received character
	return

txchar:
	btfss	PIR1,TXIF	; see if buffer register is empty
	goto	txchar		; wait until it is
	movwf	TXREG		; load character to transmit buffer regsister
	return



	;; Initialize UART for asynchronous mode at 19.2Kbps
	;;
	;; Assumes 4MHz clock
	;;
	
init:	
	bsf	STATUS,RP0
	bcf	STATUS,RP1	; data page 1
	bcf	TXSTA,SYNC	; asynchronous mode
	bsf	TXSTA,BRGH	; high speed
	movlw	0x0c
	movwf	SPBRG		; set baud rate generator
	bcf	STATUS,RP0	; data page 0
	bsf	RCSTA,SPEN	; enable serial port
	bsf	RCSTA,CREN	; enable continuous receive
	bsf	STATUS,RP0	; data page 1
	bcf	TXSTA,TX9	; only 8 bits
	bsf	TXSTA,TXEN	; enable transmitter
	bcf	TRISB,4     ; Make RB2/TX data an output
	BSF	TRISB,2		; Make sure RB1/RX is an input
	bcf	STATUS,RP0	; data page 0
	return


	end
