	include p16f628.inc
	include data.inc
	include adc.inc

	global	init_serialee,  write_inflight, write_preflight
	global	read_block, write_it_now, ee_end_preflight

#define EE_DATA	PORTA,4
#define EE_CLK  PORTB,4
#define EE_DATA_TRIS TRISA,4

EESlaveWrite	equ	0xA0
EESlaveRead	equ	0xA1
	


	code
	
	;; Code to store data in 24LC256 serial EEPROM using I2C interface.
	;; Hacked from C code of Robert Dehate. Robert directly toggled the
	;; data output line high and low rather than use the pull up resistor.
	;; I have mangled the code to make the output bit a pseudo open 
	;; drain output. The port output bit is set to a zero and then the
	;; output is toggled from high (change to input mode, ie. high-Z) to
	;; low (output mode) using the TRISA register.
	;;
	;; Prelaunch data is stored in the first 256 bytes of the EEPROM.
	;; This is written continuously during pre-launch. The EEPROM is
	;; specified for 1,000,000 erase/write cycles so assuming we store
	;; 32 SPS and 8 bytes/sample, the EEPROM is good for almost 12 days.
	;; Not good for production code but OK for testing.
	;;
	;; Once we transition to in-flight, a 4 byte marker is written and
	;; then the write address is shifted to 0x100 for in-flight data.
	;;
	;; Serial link is bit banged and timing is for a 4MHz clock
	;;

init_serialee:
	bsf	STATUS,RP0		; data page 1
	bcf	TRISB,4		; make I2C clock output an output
	bcf	STATUS, RP0		; data page 0 again
	bcf	EE_CLK
	clrf	EE_ADDRESSB0
	clrf	EE_ADDRESSB1
	return
	
ee_end_preflight:
	movlw	0x07		; adjust EE write address for start of flight
	movwf	EE_ADDRESSB0
	movlw	0xf0
	movwf	EE_ADDRESSB1
	return

data_low:
	bsf	STATUS,RP0	; data page 1
	bcf	EE_DATA_TRIS
	bcf	STATUS,RP0	; data page 0
	bcf	EE_DATA		; make sure output is low
	return

data_high:	
	bsf	STATUS,RP0	; data page 1
	bsf	EE_DATA_TRIS
	bcf	STATUS,RP0	; data page 0
	return

	;; Send I2C start bit. This requires making the data line
	;; transition to a low state while the clock line is high.
startbit:
	call	data_high	; data high
	bcf		EE_CLK			; clock low
	nop
	bsf	EE_CLK			; clock high
	call	data_low	; data low
	bcf	EE_CLK			; clock low
	call	data_high	; data line back to high-Z state
	return
	
	;; Send I2C stop bit. This requires making the data line
	;; transition to a high state while the clock line is high

stopbit:

	call	data_low	; be sure data line is an output
	nop
	bsf	EE_CLK		; clock high
	call	data_high		; data high
	bcf	EE_CLK		; clock low
	return

outputbit:
	bsf	EE_CLK		; clock high
	bcf	EE_CLK		; clock low
	call	data_high		; data high (float)
	return

	;; Write a byte of data to the serial EEPROM.
	;;
	;; Data to write is W and saved to T0
	;;
	;; The acknowledge bit is received from the EEPROM
	;;
writebyte: movwf	T0
	movlw	8
	movwf	T1			; T1 used as counter
wbloop:	btfss	T0,7		; test bit
	call	data_low	; it was a zero
	rlf	T0,F			; rotate data left one bit
	bsf	EE_CLK			; clock high
	bcf	EE_CLK			; clock low
	bsf	STATUS,RP0		; data page 1
	bsf	EE_DATA_TRIS		; float output
	bcf	STATUS,RP0		; data page 0
	decfsz	T1,F		; repeat for remaining bits
	goto	wbloop
	nop
	;; Get ACK bit
	bsf	EE_CLK		; clock high
	bcf	EE_CLK		; clock low
	return

	;; Send write command and start address

eewrite:
	call	startbit	; send start
	movlw	EESlaveWrite
	call	writebyte
	movf	EE_ADDRESSB0,W	; get high byte of EEPROM address
	call	writebyte
	movf	EE_ADDRESSB1,W	; get low byte of address
	call	writebyte
	return


	;; Store information in serial EEPROM
	;;
	;; There are three different actions that take place repending on
	;; time.
	;; During pre-launch, the data is stored in a ring buffer in
	;; the first 256 bytes of the EEPROM.
	;; At launch detect, the ring buffer is marked with 0xDEAD 0xBEEF
	;; During the rest of the flight (until space runs out) data is
	;; stored directly to the EEPROM.
	;;
	;; The EEPROM will do multi byte writes but these cannot cross
	;; 64 byte boundries. To make things easier, each sample is
	;; 16 bytes. Thus the boundry
	;; will never be crossed. Because of the limited storage space,
	;; data is stored at a variable rate. 64 times per second from launch to apogee
	;; and twice per second after that.
	;;
	;; The data stored are the state variables,
	;; raw pressure sensor reading, raw acceleration sensor reading,
	;; main state, and the LSB of the timer tick.
	;; The data is packed thus:
	;;
	;; ALTB0
	;; ALTB1
	;; ALTB2
	;; VELB0

	;; VELB1
	;; VELB2
	;; ACCELB0
	;; ACCELB1

	;; ACCELB2
	;; PRESSUREB0  (upper nybble is main state)
	;; PRESSUREB1
	;; RAWACCELB0

	;; RAWACCELB1
	;; TICKB2

	;; Note that the last two bytes are unused and filled with zeroes.

write_preflight:
	;; handle the pre-launch data and ring buffer
	;; Record two seconds of pre-launch data (128 samples)
	
	movlw	0x10		; adjust address
	addwf	EE_ADDRESSB1,F	
 	btfsc	STATUS,C
	incf	EE_ADDRESSB0,F
	movf	EE_ADDRESSB0,W
	andlw	0x07		; mask off all but three bits of upper byte
	movwf	EE_ADDRESSB0

write_data:
	swapf	MAIN_STATE,W
	iorwf	PRESSUREB0,F
	call	eewrite		; start write process

	movf	ALTB0,W
	call	writebyte
	movf	ALTB1,W
	call	writebyte
	movf	ALTB2,W
	call	writebyte

	movf	VELB0,W
	call	writebyte
	movf	VELB1,W
	call	writebyte
	movf	VELB2,W
	call	writebyte

	movf	ACCELB0,W
	call	writebyte
	movf	ACCELB1,W
	call	writebyte
	movf	ACCELB2,W
	call	writebyte


	movf	PRESSUREB0,W
	call	writebyte
	movf	PRESSUREB1,W
	call	writebyte

	movf	RAWACCELB0,W
	call	writebyte
	movf	RAWACCELB1,W
	call	writebyte

	movf	TICKB2,W
	call	writebyte


;; Debug code to read the timer 1 value
if 1

	movf	TMR1H,W
	movwf	T2
	movf	TMR1L,W
	movwf	T3
	movf	TMR1H,W
	subwf	T2,W
	btfsc	STATUS,Z
	goto	continue
	movf	TMR1H,W
	movwf	T2
	movf	TMR1L,W
	movwf	T3
continue:
	movf	T2,W
	call	writebyte
	movf	T3,W
	call	writebyte

else
	;; I have to use a total of 16 bytes (a power of two) to keep
	;; the EEPROM happy. (It doesn't cross page boundries without
	;; a lot of hassle so it pays to keep writes within a page.)
	;; I now have two bytes left over. Since the battery is connected
	;; to an ADC channel, now is the time!

	movlw	SECH3		; channel with the battery voltage
	movwf	T0
	call	readADC
	movf	T1,W		; writebyte uses T1, so copy it elsewhere for the moment
	movwf	T2
	movf	T0,W
	call	writebyte
	movf	T2,W
	call	writebyte

endif

	call	stopbit		; stop write process

	return


	;; Write the inflight data. 
	;; 
write_inflight:

	btfss	EE_ADDRESSB0,6	; see if in first half of EEPROM
	goto	write_it_now	; yes, record the data

	movf	TICKB2,w		; check time, only record twice a second
	andlw	0x1f			; this will be zero every 32 clock ticks
	btfss	STATUS,Z
	return

write_it_now:
	btfsc	EE_ADDRESSB0,7	; If msb is set, then we have overflowed the EEPROM
	return

	movlw	0x10
	addwf	EE_ADDRESSB1,F	; bump address
	btfsc	STATUS,C
	incf	EE_ADDRESSB0,F
	btfsc	EE_ADDRESSB0,7	; If msb is set, then we have overflowed the EEPROM
	return

	goto	write_data		; else, write data

	;; Read a single byte from the serial EEPROM. An acknowledge
	;; bit is is NOT sent.

readbyte:
	call	data_high	; set port for input
	clrf	T0
	bcf	STATUS,C
	movlw	8
	movwf	T1		; used for loop counter
rloop:	bsf	EE_CLK		; make clock output high
	rlf	T0,F
	btfsc	EE_DATA
	incf	T0,F
	bcf	EE_CLK		; clock output low
	decfsz	T1,F
	goto	rloop
	movf	T0,W
	return
	
	;; Read a block of 8 bytes from the serial EEPROM.
	;; Address should already be set in EE_ADDRESS and FSR should
	;; point to first (lowest) address of file memory to store data.
	;;
	
read_block:
	call	eewrite		; send address
	call	startbit	; terminate write operation with a start bit after loading address
	movlw	EESlaveRead	; and send read command
	call	writebyte
	movlw	8
	movwf	LOOPCOUNT
rbloop:	call	readbyte	; get a byte
	movwf	INDF
	incf	FSR,F
	decfsz	LOOPCOUNT,F
	goto	ack_bit
r_end:	call	stopbit		; send a stop bit
	return
ack_bit:
	call	data_low
	call	outputbit	; send ack bit
	goto	rbloop		; repeat


	end
